# Copyright(c) 2004-2006 Association of Universities for Research in Astronony, Inc.

###############################################################################
# Gemini GNIRS data reduction script                                          #
# Reduction for: XD Calibrations                                              #
#                                                                             #
# This script is provided to guide the GNIRS data user through the reduction  #
# process.  It is not intended to be an executable task, or a "plug-n-play"   #
# batch job.  It does show the reduction steps using actual GNIRS data and    #
# provides explanatory comments at each step.                                 #
#                                                                             #
#                                                                             #
# Once properly edited, the script can be run by cutting and pasting the      #
# text into an IRAF session, or by typing                                     #
#   gn> task $thisscript="thisscript.cl"                                      # 
#   gn> thisscript                                                            #
# in the IRAF session.                                                        #
#                                                                             #
# WARNING: Certain INTERACTIVE steps in the process MAY NOT WORK WELL         #
#          with the 'cl < thisscriptname.cl' syntax).  Likewise, when         #
#          cutting/pasting, DO NOT paste commands directly after an           #
#          interactive task in one paste (that is paste the following         #
#          command after the interactive task).                               #
#                                                                             #
# Note that this script is designed to be re-run as needed, and so each step  #
# is preceeded by a command to delete the files created in that step.         #
###############################################################################
#
# Calibration Dataset:
#
# Observation UT date : 2006-may-07
# Data filename prefix: S20060507S
# File numbers:
#	IR lamp flats:  128-136
#	QH lamp flats:  138-146
#	Argon arcs:     70-71
#	Pinhole: 	125
#	Telluric:	75-82
#
# The data used in this script are available in the Gemini Science Archive.   


######################################################################
		### Initialize tasks ###
######################################################################
# Load required tasks/packages
gemini
gnirs

# set default parameters
# use gnirs defaults except where specified on command lines below
unlearn gemini
unlearn gemtools
unlearn gnirs


######################################################################
		### define logfile and database ###
######################################################################

# to start from scratch, delete existing log file
#delete "gnirs_xd32_cal.log" ver- >& dev$null
gnirs.logfile = "gnirs_xd32_cal.log"

# set database directory
#to start from scratch, delete existing contents
#if (access("gnirs_xd32_database")) 
#  delete gnirs_xd32_database/*
#;
gnirs.database = "gnirs_xd32_database"

# load default header keywords for gnirs and set display
nsheaders gnirs
set stdimage=imt1024



######################################################################
		### Create lists and path to data ###
######################################################################

#set data path

set rawdir="/net/sabrina/staging5/gemini_testdata/gnirs_testdata/xdisp/"

delete *.list ver- >& dev$null

##Note: this example dataset uses the GCAL NIR balance filter for flatfields
## and so has only two sets of flats.  Flats taken before ~April 2006
## or any flats using the GCAL ND filters will have three sets of flats.
## Commands for three sets of flats are included below (but commented out)
## and called "Old flats".

#- IR lamp: for order 3 
gemlist S20060507S 128-136 > flat1.list

#- QH lamp: for orders 4-8
gemlist S20060507S 138-146 > flat2.list

## For older flats, the short exposure QH lamps are for order 4-5
## and the longer exposure QH lamps are for order 6-8
#- QH lamp long exposures: for orders 6-8
#gemlist S2006... xx-xx > flat3.list

gemlist S20060507S 125 > pinhole.list
gemlist S20060507S 70-71 > arc.list
gemlist S20060507S 75-82 > telluric.list

cat flat*.list > allflats.list
concat allflats.list,pinhole.list,arc.list,telluric.list > all.list



######################################################################
		### nsprepare the data ###
######################################################################

# Examine all data visually first.  If data show a strong vertical striping
# pattern (due to variable bias problem on readout), run nvnoise on the raw
# data to correct this.  This task subtracts a constant from each pixel
# and so does not add any noise.
#
## If nvnoise is run, all subsequent files will have a 'v' preceeding the 
## file prefix (by default).
#
#imdelete v@all.list ver- >& dev$null
#nvnoise @all.list
#

# run nsprepare to correct for non-linearity, determine MDF shifts,
# fix headers and create variance and data quality planes.  If a flat 
# is included in the file list and shiftx and shifty = INDEF, nsprepare 
# will determine the MDF shift from the flat and insert in all file headers.  
# Use fl_forcewcs=yes to correct the WCS keywords in the headers.
# Bad pixel mask added to GNIRS IRAF package with version 1.9.

imdelete n@all.list ver- >& dev$null
nsprepare @all.list rawpath=rawdir$ shiftx=INDEF shifty=INDEF fl_forc+ \
    bpm="gnirs$data/gnirs_2005sep24_bpm.fits"

## Note that nsprepare looks for a well exposed order 3 to obtain the shift,
## so shiftx/y=INDEF will work best with IR lamp flats (nominally the first 
## flat set).  If running nsprepare on a dataset that does not include a 
## flat, use the "shiftimage" parameter to match the shift to an nsprepare-d
## flat image.

##One can check the results of the cutting procedure by cutting one flat 
##image with nscut and viewing the results with nxdisplay, or by displaying 
##one of the output images from the nsreduce call below.
#

######################################################################
		### Generate flatfields ###
######################################################################


## Old set of flats: obtained before ~April 2006, or any using GCAL ND filters 
## flat1.list is IR lamp flats, good for order 3 and possibly order 4
## flat2.list is QH lamp, ok for order 4 and order 5
## flat3.list is QH lamp w/longer exposure-- orders 3-5 saturated, good for orders 6-8

##New set of flats: obtained after March 2006, using GCAL NIR Balance filter 
## Two flat sets:
## flat1.list = IR lamp flats, for order 3 (QH lamp has K-band absorption feature)
## flat2.list = QH lamp, good for orders 4-8
##

imdelete rn@allflats.list ver- >& dev$null
nsreduce n@allflats.list fl_sky- fl_nscut+ fl_flat- fl_dark- fl_nsappwave- \
    fl_corner+

## print imstatistics on some flat orders to check for anomalous mean or 
## standard deviation.  Often the first flat in a set should be discarded.
## Note: these steps create and delete a file called "tmpflat"
## Prints to screen and to log file

printlog ("-------------------------------------------- ", gnirs.logfile, ver+)
delete tmpflat ver- >& dev$null

gemextn rn@flat1.list proc="expand" extname="SCI" extver="1" > tmpflat
printlog ("Order 3 Flats: ", gnirs.logfile, ver+)
imstatistic @tmpflat | tee (gnirs.logfile, append+) 
delete tmpflat ver- >& dev$null

gemextn rn@flat2.list proc="expand" extname="SCI" extver="3" > tmpflat
printlog ("Order 5 Flats: ", gnirs.logfile, ver+)
imstatistic @tmpflat | tee (gnirs.logfile, append+)
delete tmpflat ver- >& dev$null

## Old flats (3 sets):
#gemextn rn@flat3.list proc="expand" extname="SCI" extver="4" > tmpflat
## OR New flats (2 sets):

gemextn rn@flat2.list proc="expand" extname="SCI" extver="4" > tmpflat
printlog ("Order 6 Flats: ", gnirs.logfile, ver+)
imstatistic @tmpflat | tee (gnirs.logfile, append+)
delete tmpflat ver- >& dev$null

### Run nsflat ###

## Flat names:  intermediate flats are called flat1, flat2, flat3
## final flat is named final_flat. This is the only one that will be archived.

#################### NOTES ON NSFLAT FOR XD DATA ###################
## IMPORTANT:  It is recommended to use process="fit" for XD data 
##             ("auto" also defaults to "fit" as of GEMINI v1.9).  
##             process="fit" gives a much better final flat ("trace" 
##             does not work adequately).
##
## To fit the normalization manually, set fl_inter+; to define the lower
## and upper thresholds for good pixels interactively, set fl_range+
##
## S/N given in the far right column by nsflat is incorrect for XD 
## flats.  S/N in the middle column (under "Lamps") is more accurate.
## 
## A Warning that "DQ for flat is poor" is commonly seen with XD data.
## This is because all of the un-illuminated pixels are set to "bad".
## You can ignore this warning (but check that the final flat looks good 
## and there is real signal within the illuminated regions of each order).
####################################################################

##  flat1: Order 3 ##
## If done interactively, all 8 orders will be displayed, but only 
## order 3 is used later.  Other orders can be ignored (for orders>4 
## there is almost no signal).

delete flat1.fits,flat1_bpm.pl ver- >& dev$null

# The following parameters should work well non-interactively.  An order=10
# fit is needed to fit the spectral signature of the lamp in the K-band.

nsflat rn@flat1.list flatfile="flat1.fits" fl_inter- fl_corner+ \
   process="fit" fitsec="MDF" order=10 lthresh=100. thr_flo=0.35 thr_fup=1.5
display flat1.fits[sci,1] 1 zr+ zs-


##  flat2: Orders 4-5 (old flats) OR 4-8 (new flats) ##
delete flat2.fits,flat2_bpm.pl ver- >& dev$null

# For the higher orders, better to fit each curve interactively (fl_inter+).
# A high value for the upper threshold limit (thr_fup=4.0) is needed to prevent 
# many good pixels from exceeding the limit (the calculated mean is too low).
# Alternatively, set fl_range+ to set the good pixel limits by hand.

## WARNING: use fl_inter+ here.  Paste this by itself on the command line!! ##
nsflat rn@flat2.list flatfile="flat2.fits" fl_inter+ fl_corner+ \
   process="fit" fitsec="MDF" order=5 lthresh=50. thr_flo=0.35 thr_fup=4.0

nxdisplay flat2 2 


##  flat3: Orders 6-8 (old flats only) ##
#delete flat3.fits,flat3_bpm.pl ver- >& dev$null
## WARNING:fl_inter+ here.  Paste this by itself on the command line!! ##
#nsflat rn@flat3.list flatfile="flat3.fits" fl_inter+ fl_corner+ \
#   process="fit" fitsec="MDF" order=5 lthresh=50. thr_flo=0.35 thr_fup=4.0

#nxdisplay flat3.fits 3 zr+ zs-
#

## New flats
#- create the final flat: order: 3     -> flat1 -> position 1-3, (sci: 1)
#-				 4-8   -> flat2 -> position 4-18 (sci: 2-6)

## Old flats 
#- create the final flat: order: 3     -> flat1 -> position 1-3,  (sci: 1)
#-				 4-5   -> flat2 -> position 4-9,  (sci: 2,3)
#-                               6-7-8 -> flat3 -> position 10-18 (sci: 4,5,6)

## Note about Extensions: The fxinsert commands below use extensions numbers, 
## not names, and so assume a certain MEF configuration.  The assumed 
## configuration is the one created by this example: 6 extracted orders with 
## variance and data quality planes, and no attached MDF.  It is a good idea 
## to use gemextn to verify the extensions in your generated flat files before 
## using the fxinsert command to create the final flat.  
## The output from gemextn should match the following. 
#
#gemextn flat1
#flat1[0]
#flat1[1][SCI,1]
#flat1[2][VAR,1]
#flat1[3][DQ,1]
#flat1[4][SCI,2]
#flat1[5][VAR,2]
#flat1[6][DQ,2]
#flat1[7][SCI,3]
#flat1[8][VAR,3]
#flat1[9][DQ,3]
#flat1[10][SCI,4]
#flat1[11][VAR,4]
#flat1[12][DQ,4]
#flat1[13][SCI,5]
#flat1[14][VAR,5]
#flat1[15][DQ,5]
#flat1[16][SCI,6]
#flat1[17][VAR,6]
#flat1[18][DQ,6]


## NOTE: the current (v1.9 or earlier) MDF file for XD data extracts 6 orders 
## (3-8).  Order 9 is not extracted; to add order 9, modify the MDF file in 
## gnirs$data or the MDF file attached to each data file in nsprepare.

imdelete final_flat.fits ver- >& dev$null
fxcopy flat1.fits final_flat.fits "0-3" new_file+
fxinsert flat2.fits final_flat.fits[3]    groups="4-9"
fxinsert flat2.fits final_flat.fits[9]    groups="10-18"
#OR for Old flats:
#fxinsert flat3.fits final_flat.fits[9]    groups="10-18"


reset stdimage=imt1024
nxdisplay final_flat   1

##Order 8 (sci extension 6) should flatfield properly with new flats in v1.9.
##However, if you are not happy with it (or any of the orders), you can
##use these lines to blank (to 1) any order (order 6 in this example).
#imreplace final_flat.fits[sci,5] 1.
#imreplace final_flat.fits[var,5] 0.
#imreplace final_flat.fits[dq,5] 0.
#
#imreplace final_flat.fits[sci,6] 1.
#imreplace final_flat.fits[var,6] 0.
#imreplace final_flat.fits[dq,6] 0.


######################################################################
		### Reduce Arcs & Pinholes ###
######################################################################


## - Reduce Arcs - ###
imdelete rn@arc.list ver- >& dev$null
nsreduce n@arc.list fl_nscut+ fl_nsappwave- fl_dark- fl_sky- fl_flat- \
   fl_corner+

## - Reduce Pinhole - ###
imdelete rn@pinhole.list ver- >& dev$null
nsreduce n@pinhole.list fl_nscut+ fl_nsappwave- fl_dark- fl_sky- fl_flat- \
   fl_corner+
#rename pinhole to generic name for use below (assumes only 1 pinhole image)
imdelete pinhole.fits ver- >& dev$null
imrename rn@pinhole.list pinhole


#############################
## Obtain the S-distortion ##
#############################

## This is used by nstransform later for spectral rectification.  It can
## also be used by nswavelength if 'fl_median+', but this is not recommended.  
## This command is best run interactively.

# When presented with the cross section, hit "m" on the 5 strongest peaks
# and accept the pixel values given. Do not use the faint peak to 
# the left (it's partially vignetted by the decker). For each order, 
# check if the offset between the peak and the listed value is consistent 
# for all features, and if the automated finding does not associate the 
# same value for two features.  The offset will be different for different 
# orders, and the way to override a value is to give a "better guess".
# Note the use of nlost=0 to stop the task where the pinhole ends
# this is needed to avoid "tracing into the noise" in the higher orders

## WARNING: fl_inter+ here.  Paste this by itself on the command line!! ##
nssdist pinhole coordlist=gnirs$data/pinholes-short-dense.lis \
   section="[*,100]" fl_inter+ function="legendre" order=5  minsep=5 \
   thresh=1000 nlost=0.

####################################
## Obtain the Wavelength solution ##
####################################

# Only one arc image is needed; if multiple are taken, you can combine them to
# improve signal to noise.  Using nscombine (even though there is no
# shift, because nsstack does not have an option to name output image).

imdelete arc_comb ver- >& dev$null
nscombine rn@arc.list output="arc_comb"

# With this data set, the arcs were taken with a wide slit (0.675 arcsec),
# therefore the RMS are fairly large (~1-3angstroms).  More accurate results
# can be obtained with a 2-pixel (0.3 arcsec) slit.

# Accurate line identification is usually obtained automatically for orders 
# 3-7; order 8 often needs to be corrected.  All orders should be checked 
# interactively.

imdelete warc_comb ver- >& dev$null
nswavelength arc_comb coordlist="gnirs$data/lowresargon.dat" fl_median+ \
   fl_inter+ refit- threshold=300. nlost=10 fwidth=5.

##Note on nswavelength: fl_median (new in v1.9) creates a reference arc
##for XD data by median filtering the arc in each order.  This option is
##possible because the arc lines are parallel to the rows, and avoids
##calling nstransform in nswavelength to spatially rectify the arc.  
##If spatial rectification is desirable for some reason, set fl_median=no, 
##but be sure to apply the same spatial transformation here as is used 
##later to rectify the telluric and science data.

######################################################################
		### Telluric standard ###
######################################################################

## There are 3 ways to determine the correct frames for sky subtraction
## in nsreduce.  If the frames are evenly spaced in time (i.e., ABBA 
## dither pattern), the easiest is "skyrange=INDEF" (the default), which
## will determine the best sky frame for each image from the input list.
## One can also set the skyrange (in seconds) by hand, or provide a list
## of sky images directly.  See nsreduce help for more information.
## If two or more images meet the sky selection criteria (skyrange and 
## nodsize), they will be combined to create the sky frame that is subtracted.


### - Reduce telluric ###
imdelete rn@telluric.list ver- >& dev$null
nsreduce n@telluric.list fl_corner+ fl_nsapp- fl_sky+ skyrange=INDEF \
   fl_flat+ flatimage="final_flat.fits" 

### - Combine telluric ###
# If "output" is not provided, nscombine will name the output image 
# to the first file in the list with the suffix "_comb".  Here we
# rename to a generic name for use below.

imdelete tell_comb.fits ver- >& dev$null
nscombine rn@telluric.list output="tell_comb"


### - nstransform telluric ###

## NSTRANSFORM is used to apply the wavelength solution and an
## sdistortion correction, if desired.  It can be run with either
## lamp or sdist or both.  It is best run interactively.
## If only the lamp solution is applied, the spectra will not be 
## "straightened", but they will be flipped in y because of the wavelength
## solution.  To also straighten, add the pinhole (sdisttrans="pinhole").

## The interactive mode of nstransform puts you into fitcoords, which
## plots the fit and the residuals for the 2-dimensional xy fit for each
## transformation (lamp and sdist).  See help pages for nstransform, fitcoords 
## and transform for more information.  For the lamp solution, xorder=1 
## and yorder=3 works best, and outlying points should be deleted.  For the 
## sdist (pinhole) solution, use xorder=4 and yorder=4 for XD32 data (lower 
## orders, 2-3, for XD111 data); usually the first y point and the last few 
## y points in each order should be deleted.  (Hit 'f' to redo the fit.)  If 
## the pinhole trace is lost, which may happen for the higher orders (especially 
## pre-July 2005 data), more upper y datapoints may need to be deleted.

## NOTE that it is important to apply the same solution for the telluric and
## the science data.

imdelete ttell_comb.fits ver- >& dev$null
nstransform tell_comb.fits lamptrans="warc_comb" sdisttrans="pinhole" \
   fl_inter+ fl_stripe- xorder=4 yorder=4

nxdisplay ttell_comb.fits 1


### nsextract telluric for 1D spectrum ###

imdelete xttell_comb.fits ver- >& dev$null
nsextract ttell_comb.fits line=750 nsum=20 upper=6 low=-6 fl_inter+ \
   fl_apall+ fl_trace+


######################### FINISHED  ##########################################  


## delete extra files
## keep these lines commented until you are done
#delete *list
#delete tmp*
#imdelete nS20060507S????.fits ver- >& dev$null
#imdelete rnS20060507S????.fits ver- >& dev$null


