# Cpyright(c) 2006 Association of Universities for Research in Astronomy, Inc.
#
# NFFITCOORDS -- This task takes arc and s-distortion calibration images,
# which have previously been traced, and fits dispersion and s-distortion
# surface functions.  The fits are stored in the database directory in a
# file for each input 2D spectrum.  Note that the fitting is redundant
# because the fits for many spectra will use the same calibration data
# but the overhead is low and allows unique surface functions for each
# image.  The resulting functions are recorded in the image headers for
# later use; e.g. by NFTRANSFORM and NIFCUBE.

procedure nffitcoords (inimages) 

# Determine 2D wavelength and S-distortions functions
#
# Version:Sep 26, 2005 FV  separated fitcoords and transform to separate tasks

char	inimages	{prompt = "Input NIFS spectra"}							# OLDP-1-primary-single-prefix=t
char	outimages	{"", prompt = "Output NIFS spectra"}						# OLDP-1-output
char	outprefix	{"f", prompt = "Prefix for output spectra"}					# OLDP-4
char	lamptransf	{"", prompt = "Name of arc lamp spectrum used to determine wavelength calibration"}			# OLDP-1-input
char	sdisttransf	{"", prompt = "Name of image used to determine S-distortion calibration"}	# OLDP-1-input
int	dispaxis	{1, min = 1, max = 2, prompt = "Dispersion axis if not defined in the header"}	# OLDP-3
char	database	{"", prompt = "Directory for calibration files"}				# OLDP-3
bool	fl_inter	{no, prompt = "Examine fitted transformations interactively"}			# OLDP-4
char	function	{"chebyshev", min = "legendre|chebyshev|spline1|spline3", prompt = "Coordinate fitting function"}	# OLDP-2
int	lxorder		{2, min = 1, prompt = "X order of lamp fitting function"}			# OLDP-2
int	lyorder		{2, min = 1, prompt = "Y order of lamp fitting function"}			# OLDP-2
int	sxorder		{2, min = 1, prompt = "X order of S-distortion fitting function"}			# OLDP-2
int	syorder		{2, min = 1, prompt = "Y order of S-distortion fitting function"}			# OLDP-2
char	logfile		{"", prompt = "Logfile"}							# OLDP-1
bool	verbose		{yes, prompt = "Verbose"}							# OLDP-2
int	status		{0, prompt = "Exit status (0=Good)"}						# OLDP-4
struct	*scanfile1	{"", prompt = "Internal use"}							# OLDP-4

struct	*scanfile2	{"", prompt = "Internal use"}							# OLDP-4

begin
        char	l_inimages = ""
        char	l_outimages = ""
        char	l_outprefix = ""
	char	l_lamptransf = ""
	char	l_sdisttransf = ""
	int	l_dispaxis
	char	l_database = ""
	bool	l_fl_inter
	char	l_function = ""
	int	l_lxorder
	int	l_lyorder
	int	l_sxorder
	int	l_syorder
	char	l_logfile = ""
	bool	l_verbose

        char	l_key_dispaxis = ""
        char	l_sci_ext = ""
	char	l_var_ext = ""
	char	l_dq_ext = ""

        bool	intdbg
        int	nlamps, nsdist, nver, version, nspatial, ndisp
	int	xo, yo, nx, ny, nxx, nyy
        int	nbad, nfiles, junk, nfail
	char	badhdr, name, keyfound, s_database
        char	l_fitname, l_fitname_lamp, l_fitname_sdist
        char	lampsec, sdistsec, lampsec2, sdistsec2, line
        char	img, imgin, imgout, lamp, sdist, sec, phu, key, sec2, sec4
	char	tmpin, tmplog
	real	dx, dy, x1, x2, y1, y2, wmin, wmax, smin, smax
	struct 	l_date, sline
        
        # Define temporary files
        tmpin = mktemp ("tmpin") 
        tmplog = mktemp ("tmplog") 

	# Cache parameter files.
	cache ("gemextn", "nfcheck", "nfckcal", "gemdate")
        
        # Initialize
        status = 1
        intdbg = no

	junk = fscan (	inimages, l_inimages)
	junk = fscan (	outimages, l_outimages)
	junk = fscan (	outprefix, l_outprefix)
	junk = fscan (	lamptransf, l_lamptransf)
	junk = fscan (	sdisttransf, l_sdisttransf)
	junk = fscan (	database, l_database)
	junk = fscan (	function, l_function)
	junk = fscan (	logfile, l_logfile)

	l_dispaxis =	dispaxis
	l_fl_inter =	fl_inter
	l_lxorder =	lxorder
	l_lyorder =	lyorder
	l_sxorder =	sxorder
	l_syorder =	syorder
	l_verbose =	verbose

	badhdr = ""
	junk = fscan (	nsheaders.key_dispaxis, l_key_dispaxis)
	if ("" == l_key_dispaxis) badhdr = badhdr + " key_dispaxis"
	junk = fscan (	nsheaders.sci_ext, l_sci_ext)
	if ("" == l_sci_ext) badhdr = badhdr + " sci_ext"
	junk = fscan (	nsheaders.var_ext, l_var_ext)
	if ("" == l_var_ext) badhdr = badhdr + " var_ext"
	junk = fscan (	nsheaders.dq_ext, l_dq_ext)
	if ("" == l_dq_ext) badhdr = badhdr + " dq_ext"

        if (l_logfile == "") {
            junk = fscan (nifs.logfile, l_logfile)
            if (l_logfile == "") {
                l_logfile = "nifs.log"
                printlog ("WARNING - NFFITCOORDS:  Both nftransform.logfile \
		    and nifs.logfile", l_logfile, verbose+) 
                printlog ("                        are empty.  \
		    Using " // l_logfile, l_logfile, verbose+) 
            }
        }

        printlog ("-------------------------------------------------------------------------------", l_logfile, l_verbose) 
        date | scan (l_date) 
        printlog ("NFFITCOORDS -- " // l_date, l_logfile, l_verbose) 
        printlog ("", l_logfile, l_verbose) 
        
        if (l_database == "") {
            junk = fscan (nifs.database, l_database)
            if (l_database == "") {
                l_database = "database"
                printlog ("WARNING - NFFITCOORDS:  Both nftransform.database \
		    and nifs.database", l_logfile, verbose+) 
                printlog ("                        are empty.  \
		    Using " // l_database, l_logfile, verbose+) 
            }
        }

        s_database = ""
        if (l_database != "") {
	    s_database = l_database
	    if (strldx ("/", s_database) != strlen (s_database)) {
		s_database = s_database // "/"
	    }
	}

        
	if ("" != badhdr) {
            printlog ("ERROR - NFFITCOORDS: Parameter(s) missing from \
		nsheaders: " // badhdr, l_logfile, verbose+) 
            goto clean
	}

	# Expand and verify input.
	nfcheck (input=l_inimages, output=l_outimages, prefix=l_outprefix,
	    outlist=tmpin, name="NFFITCOORDS", dependflag="*PREPAR*",
	    procflag="NFFITCOO", sci_ext=l_sci_ext, logfile=l_logfile,
	    verbose=l_verbose, vverbose=intdbg)
	if (nfcheck.status == 1)
	    goto clean
	nfiles = nfcheck.nver
	nver = nfcheck.nver

	# Verify calibrations.
	nfckcal (lampfiles=lamptransf, sdistfiles=sdisttransf,
	    outlist=tmpin, nfiles=nfiles, nver=nver, sci_ext=l_sci_ext,
	    database=s_database, name="NFFITCOORDS", logfile=l_logfile,
	    verbose=l_verbose, vverbose=intdbg)
	if (nfckcal.status == 1)
	    goto clean
	nlamps = nfckcal.nlamps
	nsdist = nfckcal.nsdist
        
        # Main loop for transformation
        scanfile1 = tmpin
        while (fscan (scanfile1, imgin, imgout, lamp, sdist) != EOF) {
	    if (intdbg) print ("processing " // imgin)

	    # If no output image requested then need to fix scan.
	    # Otherwise we need to make a copy of the input.
	    if (l_outimages == "" && l_outprefix == "") {
	        sdist = lamp; lamp = imgout; imgout = imgin
	    }

	    l_dispaxis = dispaxis
	    nfail = 0
	    wmin = INDEF; wmax = INDEF
	    smin = INDEF; smax = INDEF
	    hselect (imgin//"[0]", l_key_dispaxis, yes) | scan (l_dispaxis)

	    for (version = 1; version <= nver; version = version + 1) {
		if (intdbg) print ("version " // version)
		sec = "_" // l_sci_ext // "_" // version // "_"
		sec2 = "[" // l_sci_ext // "," // version // "]"
		sec4 = "[" // l_sci_ext // "," // version // ",inherit]"

		lampsec = lamp // sec;   sdistsec = sdist // sec
		lampsec2 = lamp // sec2; sdistsec2 = sdist // sec2

		if (intdbg) print ("checking for files")

		if ((nlamps>0 && (!access (s_database//"id"//lampsec))) ||
		    (nsdist>0 && (!access (s_database//"id"// sdistsec)))) {
		    printlog ("WARNING - NFFITCOORDS: Incomplete \
			transform information for " // imgin // sec,
			l_logfile, verbose+) 
		    if (intdbg) {
			printf ("nlamps = %d, database = %s\n",
			    nlamps, s_database//"id"//lampsec)
			printf ("nsdist = %d, database = %s\n",
			    nsdist, s_database//"id"//sdistsec)
		    }
		    nfail = nfail + 1
		    next
		}

		if (intdbg) print ("using files")
		if (intdbg) print (imgin // sec4)
		if (intdbg) print ("dispaxis: " // l_dispaxis) 

		longslit.dispaxis = l_dispaxis
		dx = INDEF; dy = INDEF

		# Fit wavelength solution to input image
		if (nlamps>0) {
		    l_fitname_lamp = imgout // sec // "lamp"
		    if (intdbg) print ("fitting wavelength to " \
			// lampsec2 // ", " // l_fitname_lamp) 

		    delete (s_database // "fc" // l_fitname_lamp,
			verify-, >& "dev$null")

		    # Juggle file names around here
		    del (s_database // "id" // lamp, verify-, >& "dev$null")
		    copy (s_database//"id"//lampsec, s_database//"id"//lamp)

		    # Check for maximum order allowed.
		    match ("features", s_database//"id"//lamp, stop-) |
			fields ("STDIN", 2, lines="1-") |
			average (opstring="new_sample") |
			scan (x1, x2, nx)
		    if (l_dispaxis == 1) {
		        xo = max (2, min (l_lxorder, int (x1)-1))
			yo = max (2, min (l_lyorder, nx)-1)
		    } else {
		        xo = max (2, min (l_lxorder, nx)-1)
			yo = max (2, min (l_lyorder, nint(x1)-1))
		    }

		    if (intdbg) print ("fitcoords...")
		    # Work around for WCS problem.
		    junk = 2; hselect (lampsec2, "WCSDIM", yes) | scan (junk)
		    if (junk != 2)
			hedit (lampsec2, "WCSDIM", del+,
			    add-, addo-, ver-, show-, update+)
		    fitcoords (lampsec2, fitname = l_fitname_lamp,
			interactive = l_fl_inter, combine = yes,
			database = l_database, deletions = "",
			function = l_function, xorder = xo,
			yorder = yo, logfiles = tmplog,
			plotfile = "")
		    if (junk != 2)
			hedit (lampsec2, "WCSDIM", junk, add+,
			    ver-, show-, update+)
		    concatenate (tmplog, l_logfile, append+)
		    delete (s_database // "id" // lamp, verify-, >& "dev$null")

		    # Extract limits.
		    scanfile2 = tmplog
		    while (fscan (scanfile2, sline) != EOF) {
		        if (fscanf (sline, "    (%d, %d) = %g  (%d, %d) = %g",
			    junk, junk, x1, junk, junk, x2) != 6)
			    next
			if (wmin == INDEF) {
			    wmin = min (x1, x2)
			    wmax = max (x1, x2)
			} else {
			    wmin = min (wmin, x1, x2)
			    wmax = max (wmax, x1, x2)
			}
		    }
		    scanfile2 = ""; delete (tmplog, verify-)

		    if (intdbg) print ("...done")
		} else
		    l_fitname_lamp = ""


		# Fit the s-distortion solution to input image
		if (nsdist>0) {
		    l_fitname_sdist = imgout // sec // "sdist"
		    if (intdbg) print ("fitting spatial to " \
			// sdistsec // ", " // l_fitname_sdist) 

		    delete (s_database // "fc" // l_fitname_sdist,
			verify-, >& "dev$null")
		    
		    # Juggle file names around here
		    del (s_database // "id" // sdist, verify-, >& "dev$null")
		    copy (s_database//"id"//sdistsec, s_database//"id"// sdist)

		    # Check for maximum order allowed.
		    match ("features", s_database//"id"//sdist, stop-) |
			fields ("STDIN", 2, lines="1-") |
			average (opstring="new_sample") |
			scan (x1, x2, nx)
		    if (l_dispaxis == 1) {
		        xo = max (2, min (l_sxorder, nx))
			yo = max (2, min (l_syorder, nint(x1)))
		    } else {
		        xo = max (2, min (l_sxorder, nint (x1)))
			yo = max (2, min (l_syorder, nx))
		    }

		    if (intdbg) print ("fitcoords...")
		    # Work around for WCS problem.
		    junk = 2; hselect (sdistsec2, "WCSDIM", yes) | scan (junk)
		    if (junk != 2)
			hedit (sdistsec2, "WCSDIM", del+,
			    add-, addo-, ver-, show-, update+)
		    fitcoords (sdistsec2, fitname = l_fitname_sdist,
			interactive = l_fl_inter, combine = yes,
			database = l_database, deletions = "",
			function = l_function, xorder = xo,
			yorder = yo, logfiles = tmplog,
			plotfile = "") 
		    if (junk != 2)
			hedit (sdistsec2, "WCSDIM", junk, add+,
			    ver-, show-, update+)
		    concatenate (tmplog, l_logfile, append+)
		    delete (s_database // "id" // sdist, verify-, >& "dev$null")

		    # Extract limits.
		    scanfile2 = tmplog
		    while (fscan (scanfile2, sline) != EOF) {
		        if (fscanf (sline, "    (%d, %d) = %g  (%d, %d) = %g",
			    junk, junk, x1, junk, junk, x2) != 6)
			    next
			if (smin == INDEF) {
			    smin = min (x1, x2)
			    smax = max (x1, x2)
			} else {
			    smin = min (smin, x1, x2)
			    smax = max (smax, x1, x2)
			}
		    }
		    scanfile2 = ""; delete (tmplog, verify-)

		    if (intdbg) print ("...done")
		} else
		    l_fitname_sdist = ""

		# If order along spatial direction = 1 then we
		# need to fix the scale (only correcting shape, 
		# not scale)
		# But only on final fit!  If we do this for spatial
		# alone then the wavelength cal cannot be combined
		# because the x size is different.
		if (nlamps>0 && nsdist>0) {
		    if (l_dispaxis == 1 && l_syorder == 1) {
			dy = 1
			printlog ("WARNING - NFFITCOORDS: Keeping \
			    spatial (y) scale constant for " \
			    // imgin // sec, l_logfile, l_verbose) 
		    } else if (l_dispaxis == 2 && l_sxorder == 1) {
			dx = 1
			printlog ("WARNING - NFFITCOORDS: Keeping \
			    spatial (x) scale constant for " \
			    // imgin // sec, l_logfile, l_verbose) 
		    }
		}

		l_fitname = l_fitname_lamp // "," // l_fitname_sdist
		printlog ("NFFITCOORDS: Fitnames for " // imgout \
		    // ": " // l_fitname, l_logfile, l_verbose) 

		# Image size
		img = imgin // "[" // l_sci_ext // "," // version // "]"
		hselect (img, "i_naxis" // l_dispaxis, yes) \
		    | scan (ndisp)
		hselect (img, "i_naxis" // (3 - l_dispaxis), yes) \
		    | scan (nspatial)

		if (version == 1) {
		    if (1 == l_dispaxis) {
			nx = INDEF;    nxx = 1
			ny = nspatial; nyy = ny
			x1 = INDEF;    x2 = INDEF
			y1 = 1;        y2 = ny
		    } else {
			nx = nspatial; nxx = nx
			ny = INDEF;    nyy = 1
			x1 = 1;        x2 = nx
			y1 = INDEF;    y2 = INDEF
		    }
		}		    

		# Add the WCS and  FITCOORDS information to the header.
		# This is only done to the science header.  The other
		# headers will get this information their science headers.
		# Copy the input to the output if needed.

		img = imgout // "[" // l_sci_ext // "," // version // "]"
		if (!imaccess (img))
		    copy (imgin//".fits", imgout//".fits", verbose-)

		if (intdbg) print ("fitcoords sci") 
		hedit (img, l_key_dispaxis, l_dispaxis, add+,
		    verify-, show-, update+)

		nfwcs (img, mdf="!mdf_file", xmdfname="DEC", ymdfname="RA",
		    dymdfname="slitsize_x", pixscale="!pixscale",
		    xoffset="0.", yoffset="0.", keepwcs-)

		hedit (img, "FCDB", l_database, add+, verify-, show-, upd+)
		if (l_fitname_lamp != "")
		    hedit (img, "FCFIT1", l_fitname_lamp,
			add+, verify-, show-, update+)
		if (l_fitname_sdist != "")
		    hedit (img, "FCFIT2", l_fitname_sdist,
			add+, verify-, show-, update+)
		if (!isindef(x1))
		    hedit (img, "FCX1", x1, add+, verify-, show-, update+)
		if (!isindef(y1))
		    hedit (img, "FCY1", y1, add+, verify-, show-, update+)
		if (!isindef(x2))
		    hedit (img, "FCX2", x2, add+, verify-, show-, update+)
		if (!isindef(y2))
		    hedit (img, "FCY2", y2, add+, verify-, show-, update+)
		if (!isindef(dx))
		    hedit (img, "FCDX", dx, add+, verify-, show-, update+)
		if (!isindef(dy))
		    hedit (img, "FCDY", dy, add+, verify-, show-, update+)
		if (!isindef(nx))
		    hedit (img, "FCNX", nx, add+, verify-, show-, update+)
		if (!isindef(ny))
		    hedit (img, "FCNY", ny, add+, verify-, show-, update+)
	    }

	    # Set PHU keywords.
	    phu = imgout // "[0]"
	    if (!imaccess (phu))
	        copy (imgin, imgout, verbose-)
	    hedit (phu, "FCNLAMPS", nlamps, add+, ver-, show-, update+)
	    hedit (phu, "FCNSDIST", nsdist, add+, ver-, show-, update+)
	    if (l_dispaxis == 1) {
		if (!isindef(wmin))
		    hedit (phu, "FCX1", wmin, add+, verify-, show-, update+)
		if (!isindef(wmax))
		    hedit (phu, "FCX2", wmax, add+, verify-, show-, update+)
		if (!isindef(smin))
		    hedit (phu, "FCY1", smin, add+, verify-, show-, update+)
		if (!isindef(smax))
		    hedit (phu, "FCY2", smax, add+, verify-, show-, update+)
	    } else {
		if (!isindef(wmin))
		    hedit (phu, "FCY1", wmin, add+, verify-, show-, update+)
		if (!isindef(wmax))
		    hedit (phu, "FCY2", wmax, add+, verify-, show-, update+)
		if (!isindef(smin))
		    hedit (phu, "FCX1", smin, add+, verify-, show-, update+)
		if (!isindef(smax))
		    hedit (phu, "FCX2", smax, add+, verify-, show-, update+)
	    }
        
        gemdate ()
        nhedit (phu, "NFFITCOO", gemdate.outdate,
            "UT Time stamp for NFFITCOORDS", comfile="NULL", after="",
            before="", update+, add+, addonly-, delete-, verify-, show-)
        nhedit (phu, "GEM-TLM", gemdate.outdate,
            "UT Last modification with GEMINI", comfile="NULL", after="",
            before="", update+, add+, addonly-, delete-, verify-, show-)
        
        }
        
	# Completed successfully
	status = 0

clean:
        if (status == 0) {
            printlog ("", l_logfile, no) 
            printlog ("NFFITCOORDS  Exit status good", l_logfile, l_verbose) 
            printlog ("-------------------------------------------------------------\
                ------------------", l_logfile, l_verbose) 
        }
        scanfile1 = ""
        delete (tmpin, ver-, >& "dev$null") 
end
